/*
 * Implementation of A* planner (Peter E. Hart et al., 1972)
 * Copyright (C) 2009 Zhi Yan <yz@ai.univ-paris8.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * version 0.1.1
 *
 * - TODO -
 * 1) Based on the robot's onboard computing power, replace list with array.
 * 2) collision test - path is too close to obstacles !
 * 3) pathfinding is very slow if no path exists
 */

#include <cstdio>
#include <list>
#include "pngrid.hpp"

#define ABS(a) (((a) < 0) ? -(a) : (a))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

static const int EUCLIDEAN = 0;
static const int MANHATTAN = 1;
static const int CHEBYSHEV = 2;

typedef struct Astar_node {
  int x, y; // node location
  int f, g, h; // for the evaluation function: f(n) = g(n) + h(n)
  int parent_x, parent_y; // parent node
} Node;

int distBetween(Node &a, Node &b, int type) {
  if(type == EUCLIDEAN) {
    if(a.x != b.x && a.y != b.y)
      return 14;
    else
      return 10;
  }
  
  if(type == MANHATTAN) {
    return ((ABS(a.x - b.x) + ABS(a.y - b.y)) * 10);
  }
  
  if(type == CHEBYSHEV) {
    return (MAX(ABS(a.x - b.x), ABS(a.y - b.y)) * 10);
  }
  
  return -1;
}

bool compare(Node &a, Node &b) {
  return (a.f > b.f);
}

bool nodeInList(Node &node, std::list<Node> &list, std::list<Node>::iterator &it) {
  for(it = list.begin(); it != list.end(); it++)
    if(node.x == it->x && node.y == it->y)
      return true;
  
  return false;
}

bool astarPlanner(Grid &grid, Node &start, Node &goal, std::list<Node> &path) {
  std::list<Node> open_list, closed_list;
  std::list<Node>::iterator it;
  Node current_node, neighbor_node, waypoint;
  int i, j;
  int tentative_g;
  bool tentative_is_better;
  
  start.g = 0;
  start.h = distBetween(start, goal, MANHATTAN);
  start.f = start.g + start.h;
  open_list.push_back(start);
  
  while(!open_list.empty()) {
    open_list.sort(compare);
    current_node = open_list.back();
    
    // the path has been found
    if(current_node.x == goal.x && current_node.y == goal.y) {
      waypoint = closed_list.back();
      while(waypoint.x != start.x || waypoint.y != start.y) {
	path.push_back(waypoint);
	nodeInList(waypoint, closed_list, it);
	waypoint.x = it->parent_x;
	waypoint.y = it->parent_y;
      }
      return true;
    }
    
    open_list.pop_back();
    closed_list.push_back(current_node);
    
    for(j = current_node.y-1; j <= current_node.y+1; j++) {
      for(i = current_node.x-1; i <= current_node.x+1; i++) {
	neighbor_node.x = i;
	neighbor_node.y = j;
	
	if((j < 0 || j > grid.getHeight()-1) || (i < 0 || i > grid.getWidth()-1)) // grid map boundary testing
	  continue;
	if(grid.getData()[j][i] == OBST)
	  continue;
	if(i == current_node.x && j == current_node.y)
	  continue;
	if(nodeInList(neighbor_node, closed_list, it))
	  continue;
	
	tentative_g = current_node.g + distBetween(current_node, neighbor_node, EUCLIDEAN);
	tentative_is_better = false;
	
	if(nodeInList(neighbor_node, open_list, it)) {
	  if(tentative_g < it->g)
	    tentative_is_better = true;
	} else {
	  open_list.push_back(neighbor_node);
	  it = open_list.end();
	  it--;
	  tentative_is_better = true;
	}
	
	if(tentative_is_better) {
	  it->parent_x = current_node.x;
	  it->parent_y = current_node.y;
	  it->g = tentative_g;
	  it->h = distBetween(*it, goal, MANHATTAN);
	  it->f = it->g + it->h;
	}
      }
    }
  }
  // no path exists
  return false;
}

int main(int argc, char *argv[]) {
  if(argc != 2) {
    fprintf(stderr, "Usage: program_name <file.png>\n");
    return 1;
  }
  
  // initialisation of png file and grid map 
  Png png_file = readPngFile(argv[1]);
  Grid grid_map(png_file.height, png_file.width);
  
  // load png image into grid
  png2grid(png_file, grid_map);
  
  // A* algorithm
  Node start_node = {30, 470}, goal_node = {470, 30};
  std::list<Node> path_list;
  std::list<Node>::iterator it;
  astarPlanner(grid_map, start_node, goal_node, path_list);
  
  // draw path
  for(it = path_list.begin(); it != path_list.end(); it++)
    drawPixel(grid_map, it->x, it->y, 0);
  
  drawPixel(grid_map, start_node.x, start_node.y, 2);
  drawPixel(grid_map, goal_node.x, goal_node.y, 2);
  
  // // custom output image filename
  // int length = strlen(argv[1]);
  // char *filename = new char[length+7];
  // strncpy(filename, argv[1], length-4);
  // filename[length-4] = '\0';
  // strcat(filename, "_output.png");
  
  // save grid into png image
  grid2png(png_file, grid_map, strcat(argv[1], "_output.png"));
  
  return 0;
}
