/*
 * pngio - Png image input output
 * Copyright (C) 2009 Zhi Yan <yz@ai.univ-paris8.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <cstdio>
#include <cstdlib>
#include <cstdarg>
#include "pngio.hpp"

void _abort(const char *s, ...) {
  va_list args;
  va_start(args, s);
  vfprintf(stderr, s, args);
  fprintf(stderr, "\n");
  va_end(args);
  abort();
}

Png readPngFile(char *filename) {
  FILE *fp;
  Png png;
  char header[8]; // 8 is the maximum size that can be checked
  
  // open file and test for it being a png
  if(!(fp = fopen(filename, "rb")))
    _abort("[readPngFile] File %s could not be opened for reading", filename);
  
  fread(header, 1, 8, fp);
  if(png_sig_cmp((png_byte *)header, 0, 8))
    _abort("[readPngFile] File %s is not recognized as a PNG file", filename);
  
  // initialize stuff
  if(!(png.struct_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)))
    _abort("[readPngFile] png_create_read_struct failed");
  
  if(!(png.info_ptr = png_create_info_struct(png.struct_ptr)))
    _abort("[readPngFile] png_create_info_struct failed");
  
  if(setjmp(png_jmpbuf(png.struct_ptr)))
    _abort("[readPngFile] Error during init_io");
  
  png_init_io(png.struct_ptr, fp);
  png_set_sig_bytes(png.struct_ptr, 8);
  
  png_read_info(png.struct_ptr, png.info_ptr);
  
  png.width = png.info_ptr->width;
  png.height = png.info_ptr->height;
  png.color_type = png.info_ptr->color_type;
  png.bit_depth = png.info_ptr->bit_depth;
  
  png.number_of_passes = png_set_interlace_handling(png.struct_ptr);
  png_read_update_info(png.struct_ptr, png.info_ptr);
  
  // read file
  if(setjmp(png_jmpbuf(png.struct_ptr)))
    _abort("[readPngFile] Error during read_image");
  
  png.row_pointers = (png_bytep *)malloc(sizeof(png_bytep) * png.height);
  for(int i = 0; i < png.height; i++)
    png.row_pointers[i] = (png_byte *)malloc(png.info_ptr->rowbytes);
  
  png_read_image(png.struct_ptr, png.row_pointers);
  
  fclose(fp);
  
  return png;
}

void writePngFile(Png png, char *filename) {
  FILE *fp;
  
  // create file
  if(!(fp = fopen(filename, "wb")))
    _abort("[writePngFile] File %s could not be opened for writing", filename);
  
  // initialize stuff
  if(!(png.struct_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)))
    _abort("[writePngFile] png_create_write_struct failed");
  
  if(!(png.info_ptr = png_create_info_struct(png.struct_ptr)))
    _abort("[writePngFile] png_create_info_struct failed");
  
  if(setjmp(png_jmpbuf(png.struct_ptr)))
    _abort("[writePngFile] Error during init_io");
  
  png_init_io(png.struct_ptr, fp);
  
  // write header
  if(setjmp(png_jmpbuf(png.struct_ptr)))
    _abort("[writePngFile] Error during writing header");
  
  png_set_IHDR(png.struct_ptr, png.info_ptr, png.width, png.height,
	       png.bit_depth, png.color_type, PNG_INTERLACE_NONE,
	       PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
  
  png_write_info(png.struct_ptr, png.info_ptr);
  
  // write bytes
  if(setjmp(png_jmpbuf(png.struct_ptr)))
    _abort("[writePngFile] Error during writing bytes");
  
  png_write_image(png.struct_ptr, png.row_pointers);
  
  // end write
  if(setjmp(png_jmpbuf(png.struct_ptr)))
    _abort("[writePngFile] Error during end of write");
  
  png_write_end(png.struct_ptr, NULL);
  
  // cleanup heap allocation
  // for(int i = 0; i < png.height; i++)
  //   free(png.row_pointers[i]);
  // free(png.row_pointers);
  
  fclose(fp);
}
